import asyncio
from typing import Any, AsyncGenerator, Awaitable, Callable, Iterable

from gamla import functional_generic


def run_sync(f):
    """Runs a coroutine in a synchronous context, blocking until result arrives."""
    loop = asyncio.new_event_loop()
    return loop.run_until_complete(asyncio.ensure_future(f, loop=loop))


@functional_generic.curry
async def amap_ascompleted(
    f: Callable[[Any], Awaitable[Any]], it: Iterable
) -> AsyncGenerator[Any, None]:
    for future in asyncio.as_completed(map(f, it)):
        yield await future


@functional_generic.curry
async def aexcepts(exception_type, func, handler, x):
    try:
        return await func(x)
    except exception_type as error:
        return handler(error)


@functional_generic.curry
async def mapa(f, it):
    async for element in it:
        yield f(element)


async def aconcat(async_generators):
    async for g in async_generators:
        for x in g:
            yield x


def afirst(*funcs, exception_type):
    async def afirst_inner(x):
        for f in funcs:
            try:
                return await functional_generic.to_awaitable(f(x))
            except exception_type:
                pass
        raise exception_type

    return afirst_inner



import asyncio
import functools
import inspect
from typing import Callable, Tuple

import toolz
from toolz import curried
from toolz.curried import operator

from gamla import functional


def compose_left(*funcs):
    return compose(*reversed(funcs))


_any_is_async = toolz.compose(any, curried.map(asyncio.iscoroutinefunction))


async def to_awaitable(value):
    if inspect.isawaitable(value):
        return await value
    return value


def _acompose(*funcs):
    async def composed(*args, **kwargs):
        for f in reversed(funcs):
            args = [await to_awaitable(f(*args, **kwargs))]
            kwargs = {}
        return toolz.first(args)

    return composed


def compose(*funcs):
    if _any_is_async(funcs):
        composed = _acompose(*funcs)
        # TODO(uri): Far from a perfect id, but should work most of the time.
        # Improve by having higher order functions create meaningful names (e.g. `map`).
        # Copying `toolz` convention.
        composed.__name__ = "_of_".join(map(lambda x: x.__name__, funcs))
        return composed
    return toolz.compose(*funcs)


def _make_amap(f):
    async def amap_inner(it):
        return await asyncio.gather(*curried.map(f, it))

    return amap_inner


def gamla_map(*args):
    if len(args) == 2:
        f, it = args
        if asyncio.iscoroutinefunction(f):
            return _make_amap(f)(it)
        return curried.map(f, it)
    [f] = args
    if asyncio.iscoroutinefunction(f):
        return _make_amap(f)
    return curried.map(f)


map = gamla_map


@toolz.curry
def after(f1, f2):
    return compose(f1, f2)


@toolz.curry
def before(f1, f2):
    return compose_left(f1, f2)


def lazyjuxt(*funcs):
    """Reverts to eager implementation if any of `funcs` is async."""
    if _any_is_async(funcs):

        async def lazyjuxt_inner(value):
            return await toolz.pipe(
                funcs,
                curried.map(toolz.compose_left(functional.apply(value), to_awaitable)),
                functional.star(asyncio.gather),
            )

        return lazyjuxt_inner
    return compose_left(functional.apply, curried.map, functional.apply(funcs))


juxt = compose(after(tuple), lazyjuxt)
alljuxt = compose(after(all), lazyjuxt)
anyjuxt = compose(after(any), lazyjuxt)
juxtcat = compose(after(toolz.concat), lazyjuxt)


def ternary(condition, f_true, f_false):
    if _any_is_async([condition, f_true, f_false]):

        async def ternary_inner_async(*args, **kwargs):
            return (
                await to_awaitable(f_true(*args, **kwargs))
                if await to_awaitable(condition(*args, **kwargs))
                else await to_awaitable(f_false(*args, **kwargs))
            )

        return ternary_inner_async

    def ternary_inner(*args, **kwargs):
        return (
            f_true(*args, **kwargs)
            if condition(*args, **kwargs)
            else f_false(*args, **kwargs)
        )

    return ternary_inner


curried_ternary = ternary


def pipe(val, *funcs):
    return compose_left(*funcs)(val)


def _curry_helper(f, args_so_far, kwargs_so_far, *args, **kwargs):
    f_len_args = inspect.signature(f).parameters
    args_so_far += args
    kwargs_so_far = toolz.merge(kwargs_so_far, kwargs)
    len_so_far = len(args_so_far) + len(kwargs_so_far)
    if len_so_far > len(f_len_args):
        return f(*args_so_far)
    if len_so_far == len(f_len_args):
        return f(*args_so_far, **kwargs_so_far)
    if len_so_far + 1 == len(f_len_args) and asyncio.iscoroutinefunction(f):

        @functools.wraps(f)
        async def curry_inner_async(*args, **kwargs):
            return await f(
                *(args_so_far + args), **(toolz.merge(kwargs_so_far, kwargs))
            )

        return curry_inner_async

    @functools.wraps(f)
    def curry_inner(*args, **kwargs):
        return _curry_helper(f, args_so_far, kwargs_so_far, *args, **kwargs)

    return curry_inner


def _infer_defaults(f):
    params = inspect.signature(f).parameters
    kwargs = {}
    for p in params.values():
        if p.default != p.empty:
            kwargs[p.name] = p.default
    return kwargs


def curry(f):
    @functools.wraps(f)
    def indirection(*args, **kwargs):
        return _curry_helper(f, (), _infer_defaults(f), *args, **kwargs)

    return indirection


def _compose_over_binary_curried(composer):
    def composition_over_binary_curried(*args):
        f = composer(args[0])
        if len(args) == 2:
            return f(args[1])
        return f

    return composition_over_binary_curried


allmap = _compose_over_binary_curried(compose(after(all), gamla_map))
anymap = _compose_over_binary_curried(compose(after(any), gamla_map))


itemmap = _compose_over_binary_curried(
    compose(after(dict), before(dict.items), gamla_map)
)
keymap = _compose_over_binary_curried(
    compose(itemmap, lambda f: juxt(f, toolz.second), before(toolz.first))
)

valmap = _compose_over_binary_curried(
    compose(itemmap, lambda f: juxt(toolz.first, f), before(toolz.second))
)


pair_with = _compose_over_binary_curried(lambda f: juxt(f, toolz.identity))
pair_right = _compose_over_binary_curried(lambda f: juxt(toolz.identity, f))

filter = _compose_over_binary_curried(
    compose(
        after(compose(curried.map(toolz.second), curried.filter(toolz.first))),
        gamla_map,
        pair_with,
    )
)


_first_truthy_index = compose_left(
    enumerate,
    curried.filter(toolz.second),
    curried.map(toolz.first),
    toolz.excepts(StopIteration, toolz.first),
)


class NoConditionMatched(Exception):
    pass


def _case(predicates: Tuple[Callable, ...], mappers: Tuple[Callable, ...]):
    """Case with functions.

    Handles async iff one of the predicates is async.
    Raises `KeyError` if no condition matched.
    """
    return compose_left(
        pair_right(
            compose_left(
                lazyjuxt(*predicates),
                _first_truthy_index,
                functional.check(
                    toolz.complement(operator.eq(None)), NoConditionMatched
                ),
                mappers.__getitem__,
            )
        ),
        functional.star(functional.apply),
    )


def case(predicates_and_mappers: Tuple[Tuple[Callable, Callable], ...]):
    predicates = tuple(map(toolz.first, predicates_and_mappers))
    mappers = tuple(map(toolz.second, predicates_and_mappers))
    return _case(predicates, mappers)


case_dict = compose_left(dict.items, tuple, case)



import dataclasses
import json
from typing import Any, Dict, Iterable, Optional, Text, Tuple

import dataclasses_json
import frozendict
import toolz
from toolz import curried


def get_encode_config():
    return dataclasses.field(
        metadata=dataclasses_json.config(
            encoder=lambda lst: sorted(lst, key=json.dumps, reverse=False)
        )
    )


def freeze_deep(value):
    if isinstance(value, str):
        return value
    if isinstance(value, dict) or isinstance(value, frozendict.frozendict):
        return toolz.pipe(
            value,
            dict,  # In case input is already a `frozendict`.
            curried.valmap(freeze_deep),
            frozendict.frozendict,
        )
    if isinstance(value, Iterable):
        return toolz.pipe(value, curried.map(freeze_deep), tuple)
    return value


@toolz.curry
def dict_to_csv(
    table: Dict[Any, Tuple], titles: Optional[Tuple] = None, separator: Text = "\t"
) -> Text:
    return toolz.pipe(
        table,
        dict_to_tuple_of_tuples,
        tuple_of_tuples_to_csv(titles=titles, separator=separator),
    )


dict_to_tuple_of_tuples = toolz.compose_left(
    dict.items,
    curried.map(curried.compose_left(lambda x: (x[0], *x[1]), curried.map(str), tuple)),
    tuple,
)


@toolz.curry
def tuple_of_tuples_to_csv(
    tuple_of_tuples: Tuple[Tuple[Any], ...], separator: Text = "\t"
) -> Text:
    return toolz.pipe(
        tuple_of_tuples,
        curried.map(toolz.compose_left(curried.map(str), tuple, separator.join)),
        "\n".join,
    )


class Enum(frozenset):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError



import asyncio

import pytest
import toolz
from toolz.curried import operator

from gamla import functional, functional_generic

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def _opposite_async(x):
    await asyncio.sleep(1)
    return not x


def test_do_if():
    assert functional.do_if(lambda _: True, lambda x: 2)(1) == 1


def test_currying():
    @functional_generic.curry
    def f(x, y, z):
        return x + y + z

    assert f(1, 2, 3) == 6
    assert f(1)(2, 3) == 6
    assert f(1, 2)(3) == 6


def test_juxt():
    assert functional_generic.juxt(toolz.identity, lambda x: x + 1)(3) == (3, 4)


async def test_juxt_async():
    async def slow_identity(x):
        await asyncio.sleep(1)
        return x

    assert await functional_generic.juxt(toolz.identity, slow_identity)(3) == (3, 3)


def test_anyjuxt():
    assert functional_generic.anyjuxt(lambda x: not x, lambda x: x)(True)


def test_alljuxt():
    assert not functional_generic.alljuxt(lambda x: not x, lambda x: x)(True)


async def test_alljuxt_async():
    assert not await functional_generic.alljuxt(_opposite_async, toolz.identity)(True)


async def test_anyjuxt_async():
    assert await functional_generic.anyjuxt(_opposite_async, toolz.identity)(True)


async def test_anymap():
    assert functional_generic.anymap(_opposite_async, [True, True, False])


async def test_allmap():
    def opposite(x):
        return not x

    assert not functional_generic.allmap(opposite, [True, True, False])


async def test_anymap_async():
    assert await functional_generic.anymap(_opposite_async, [True, True, False])


async def test_allmap_async():
    assert not await functional_generic.allmap(_opposite_async, [True, True, False])


async def test_allmap_in_async_pipe():
    assert not await functional_generic.pipe(
        [True, True, False],
        functional_generic.allmap(_opposite_async),
        # Check that the `pipe` serves a value and not a future.
        functional.check(lambda x: isinstance(x, bool), AssertionError),
    )


async def test_anymap_in_pipe():
    assert not functional_generic.pipe(
        [True, True, False], functional_generic.allmap(lambda x: not x)
    )


async def test_itemmap_async_sync_mixed():
    assert await functional_generic.pipe(
        {True: True},
        functional_generic.itemmap(
            functional_generic.compose(tuple, functional_generic.map(_opposite_async))
        ),
        functional_generic.itemmap(
            functional_generic.compose(tuple, functional_generic.map(lambda x: not x))
        ),
    ) == {True: True}


async def test_keymap_async_curried():
    assert await functional_generic.keymap(_opposite_async)({True: True}) == {
        False: True
    }


async def test_valmap_sync_curried():
    assert functional_generic.valmap(lambda x: not x)({True: True}) == {True: False}


async def _is_even_async(x):
    await asyncio.sleep(0.1)
    return x % 2 == 0


async def test_filter_curried_async_sync_mix():
    assert await functional_generic.pipe(
        [1, 2, 3, 4],
        functional_generic.filter(_is_even_async),
        functional_generic.map(lambda x: x + 10),
        tuple,
    ) == (12, 14)


async def test_wrap_str():
    assert toolz.pipe("john", functional.wrap_str("hi {}")) == "hi john"


def test_case_single_predicate():
    assert functional_generic.case_dict({toolz.identity: toolz.identity})(True)


def test_case_multiple_predicates():
    assert not functional_generic.case_dict(
        {operator.not_: toolz.identity, toolz.identity: operator.not_}
    )(True)


def test_case_no_predicate():
    with pytest.raises(functional_generic.NoConditionMatched):
        functional_generic.case_dict(
            {operator.not_: toolz.identity, operator.not_: toolz.identity}
        )(True)


async def test_case_async():
    assert not await functional_generic.case_dict(
        {_opposite_async: toolz.identity, toolz.identity: _opposite_async}
    )(True)


async def test_partition_when():
    assert functional.partition_when(lambda x: x == 1, []) == ()

    assert tuple(
        functional.partition_when(lambda x: x == 1, [1, 1, 2, 2, 1, 1, 2, 1, 1, 1])
    ) == ((1,), (1,), (2, 2, 1), (1,), (2, 1), (1,), (1,))


async def test_drop_last_while():
    assert tuple(functional.drop_last_while(lambda x: x == 1, [])) == ()
    assert tuple(functional.drop_last_while(lambda x: x == 1, [1])) == ()
    assert tuple(functional.drop_last_while(lambda x: x == 1, [2])) == (2,)
    assert tuple(
        functional.drop_last_while(lambda x: x == 1, [1, 1, 2, 2, 1, 1, 2, 1, 1, 1])
    ) == (1, 1, 2, 2, 1, 1, 2)



from gamla import graph


def test_get_connectivity_components():
    assert list(
        graph.get_connectivity_components({1: [2], 2: [1, 3], 3: [2], 4: [5], 5: [4]})
    ) == [frozenset({1, 2, 3}), frozenset({4, 5})]


def test_cliques_to_graph():
    assert graph.cliques_to_graph([{1, 2}, {2, 3}, {7, 8}]) == {
        1: frozenset({2}),
        2: frozenset({1, 3}),
        3: frozenset({2}),
        7: frozenset({8}),
        8: frozenset({7}),
    }


def test_cycles():
    assert graph.has_cycle({1: [1]})
    assert graph.has_cycle({1: [2], 2: [3], 3: [1]})
    assert graph.has_cycle({1: [2], 2: [3], 3: [4], 4: [2]})
    assert not graph.has_cycle({1: [2]})
    assert not graph.has_cycle({1: [2], 2: [3]})


def test_bfs_no_double_visit():
    g = {"a": ["b", "d", "e"], "b": ["f"], "e": ["f"], "d": [], "f": []}
    result = list(graph.graph_traverse_many("a", g.__getitem__))

    for n in {"a", "b", "d", "e", "f"}:
        assert n in result

    assert len(result) == len(set(result))



import itertools
from typing import Any, Callable, Dict, FrozenSet, Iterable, Set, Text, Tuple

import toolz
from toolz import curried
from toolz.curried import operator

from gamla import functional, functional_generic


@toolz.curry
def graph_traverse(
    source: Any, get_neighbors: Callable, key: Callable = toolz.identity
) -> Iterable:
    yield from graph_traverse_many([source], get_neighbors=get_neighbors, key=key)


@toolz.curry
def graph_traverse_many(
    sources: Any, get_neighbors: Callable, key: Callable = toolz.identity
) -> Iterable:
    """BFS over a graph, yielding unique nodes.
    Note: `get_neighbors` must return elements without duplicates."""
    seen_set: Set = set()
    remember = toolz.compose_left(key, seen_set.add)
    should_traverse = toolz.compose_left(key, toolz.complement(seen_set.__contains__))
    yield from general_graph_traverse_many(
        sources, get_neighbors, remember, should_traverse
    )


@toolz.curry
def general_graph_traverse_many(
    sources: Any, get_neighbors: Callable, remember: Callable, should_traverse: Callable
) -> Iterable:
    """BFS over a graph, yielding unique nodes.
    Note: `get_neighbors` must return elements without duplicates."""

    queue = [*sources]
    for element in queue:
        remember(element)

    while queue:
        current = queue.pop()
        yield current
        for node in get_neighbors(current):
            if should_traverse(node):
                remember(node)
                queue = [node] + queue


def traverse_graph_by_radius(
    source: Any, get_neighbors: Callable, radius: int
) -> Iterable:
    """Like `graph_traverse`, but does not traverse farther from given `radius`."""

    def get_neighbors_limiting_radius(
        current_and_distance: Tuple[Text, int]
    ) -> Iterable[Tuple[Text, int]]:
        current, distance = current_and_distance
        if distance < radius:
            yield from map(
                lambda neighbor: (neighbor, distance + 1), get_neighbors(current)
            )

    return map(
        toolz.first,
        graph_traverse(source=(source, 0), get_neighbors=get_neighbors_limiting_radius),
    )


edges_to_graph = toolz.compose(
    curried.valmap(toolz.compose(frozenset, curried.map(toolz.second))),
    curried.groupby(toolz.first),
)

graph_to_edges = toolz.compose_left(
    curried.keymap(lambda x: (x,)),
    dict.items,
    curried.mapcat(functional.star(itertools.product)),
)

reverse_graph = toolz.compose_left(
    graph_to_edges, curried.map(toolz.compose_left(reversed, tuple)), edges_to_graph
)

cliques_to_graph = toolz.compose_left(
    curried.mapcat(lambda clique: itertools.permutations(clique, r=2)), edges_to_graph
)


def get_connectivity_components(graph: Dict) -> Iterable[FrozenSet]:
    """Graph is assumed to undirected, so each edge must appear both ways."""
    nodes_left = frozenset(graph)
    while nodes_left:
        result = frozenset(
            graph_traverse(
                source=toolz.first(nodes_left),
                get_neighbors=toolz.compose(
                    curried.filter(operator.contains(nodes_left)), graph.get
                ),
            )
        )
        yield result
        nodes_left -= result


@toolz.curry
def groupby_many(f, it):
    """Return a mapping `{y: {x s.t. y \in f(x)}}, where x \in it. `

    Parameters:
    f (Callable): Key function (given object in collection outputs tuple of keys).
    it (Iterable): Collection.

    Returns:
    Dict[Text, Any]: Dictionary where key has been computed by the `f` key function.

    >>> names = ['alice', 'bob', 'charlie', 'dan', 'edith', 'frank']
    >>> groupby_many(lambda name: (name[0], name[-1]), names)
    {'a': frozenset({'alice'}),
     'e': frozenset({'alice', 'charlie', 'edith'}),
     'b': frozenset({'bob'}),
     'c': frozenset({'charlie'}),
     'd': frozenset({'dan'}),
     'n': frozenset({'dan'}),
     'h': frozenset({'edith'}),
     'f': frozenset({'frank'}),
     'k': frozenset({'frank'})}
    """
    return toolz.pipe(
        it,
        curried.mapcat(
            toolz.compose_left(
                lambda element: (f(element), [element]),
                functional.star(itertools.product),
            )
        ),
        edges_to_graph,
    )


@toolz.curry
def groupby_many_reduce(key: Callable, reducer: Callable, seq: Iterable):
    """
    Group a collection by a key function, when the value is given by a reducer function.

    Parameters:
    key (Callable): Key function (given object in collection outputs key).
    reducer (Callable): Reducer function (given object in collection outputs new value).
    seq (Iterable): Collection.

    Returns:
    Dict[Text, Any]: Dictionary where key has been computed by the `key` function
    and value by the `reducer` function.

    """
    result: Dict[Any, Any] = {}
    for element in seq:
        for key_result in key(element):
            result[key_result] = reducer(result.get(key_result, None), element)
    return result


@toolz.curry
def _has_cycle(sourced, get_neighbors, visited, node):
    if node in sourced:
        return True
    if node in visited:
        return False
    visited.add(node)
    return toolz.pipe(
        node,
        get_neighbors,
        functional_generic.anymap(_has_cycle(sourced | {node}, get_neighbors, visited)),
    )


def has_cycle(graph):
    return toolz.pipe(
        graph,
        dict.keys,
        curried.map(_has_cycle(frozenset(), curried.get(seq=graph, default=()), set())),
        any,
    )



# flake8: noqa

from .data import *
from .functional import *
from .functional_async import *
from .functional_generic import *
from .graph import *
from .graph_async import *
from .io_utils import *
